#include <stdio.h>
#include <CL/cl.h>
#include <malloc.h>
#include <cmath>
#include <time.h>
#include <sys/time.h>

#define CHECK_CLSTATUS(status,func)                                  \
  if (status != CL_SUCCESS) {                                   \
    printf("%s status = %d, failed at line %d, exit\n", func, status, __LINE__); \
    exit(1);                                                            \
  }

#define GX 10240

char* programSource;

void readSource(const char* kernelPath) {

  cl_int status;
  FILE *fp;
  long int size;

  printf("Program file is: %s\n", kernelPath);

  fp = fopen(kernelPath, "rb");
  if(!fp) {
    printf("Could not open kernel file\n");
    exit(-1);
  }
  status = fseek(fp, 0, SEEK_END);
  if(status != 0) {
    printf("Error seeking to end of file\n");
    exit(-1);
  }
  size = ftell(fp);
  if(size < 0) {
    printf("Error getting file position\n");
    exit(-1);
  }

  rewind(fp);

  programSource = (char *)malloc(size + 1);

  int i;
  for (i = 0; i < size+1; i++) {
    programSource[i]='\0';
  }

  if(programSource == NULL) {
    printf("Error allocating space for the kernel source\n");
    exit(-1);
  }

  fread(programSource, 1, size, fp);
  programSource[size] = '\0';

  return;
}



int main() {

  int in_data1[GX];
  int in_data2[GX];
  int out_data[GX];
  for(int i=0; i < GX; i++){
    in_data1[i] = i;
    in_data2[i] = i + i % 5;
  }

  cl_int status;  
  cl_uint numPlatforms = 0;
  cl_platform_id *platforms = NULL;

  status = clGetPlatformIDs(0, NULL, &numPlatforms);
  CHECK_CLSTATUS(status, "clGetPlatformIDs")
  printf("numPlatforms = %d\n", numPlatforms);

  platforms =   
    (cl_platform_id*)malloc(
        numPlatforms*sizeof(cl_platform_id));

  status = clGetPlatformIDs(numPlatforms, platforms, 
      NULL);
  CHECK_CLSTATUS(status, "clGetPlatformIDs")

  cl_uint numDevices = 0;
  cl_device_id *devices = NULL;
	
  printf("Begin clGetDeviceIDs() \n");
  status = clGetDeviceIDs(
      platforms[0], 
      CL_DEVICE_TYPE_GPU, 
      0, 
      NULL, 
      &numDevices);
  printf("Finish clGetDeviceIDs() \n");
  CHECK_CLSTATUS(status, "clGetDeviceIDs")
  printf("numDevices = %d\n", numDevices);

  devices = 
    (cl_device_id*)malloc(
        numDevices*sizeof(cl_device_id));

  status = clGetDeviceIDs(
      platforms[0], 
      CL_DEVICE_TYPE_GPU,        
      numDevices, 
      devices, 
      NULL);
  CHECK_CLSTATUS(status, "clGetDeviceIDs")

  cl_context context = NULL;

  context = clCreateContext(
      NULL, 
      numDevices, 
      devices, 
      NULL, 
      NULL, 
      &status);
  CHECK_CLSTATUS(status, "clCreateContext")

  cl_command_queue cmdQueue;

  cmdQueue = clCreateCommandQueue(
      context, 
      devices[0], 
      0, 
      &status);
  CHECK_CLSTATUS(status, "clCreateCommandQueue")

  cl_mem in_buffer1, in_buffer2;  
  cl_mem out_buffer;  

  in_buffer1 = clCreateBuffer(
      context,
      CL_MEM_READ_ONLY,
      sizeof(in_data1), 
      NULL, 
      &status);
  CHECK_CLSTATUS(status, "clCreateBuffer")
  in_buffer2 = clCreateBuffer(
      context,
      CL_MEM_READ_ONLY,
      sizeof(in_data2), 
      NULL, 
      &status);
  CHECK_CLSTATUS(status, "clCreateBuffer")

  out_buffer = clCreateBuffer(
      context,
      CL_MEM_WRITE_ONLY,                 
      sizeof(out_data), 
      NULL, 
      &status);
  CHECK_CLSTATUS(status, "clCreateBuffer")

  readSource("ele_sub.cl");
  cl_program program = clCreateProgramWithSource(
      context, 
      1, 
      (const char**)&programSource,                                 
      NULL, 
      &status);
  CHECK_CLSTATUS(status, "clCreateProgramWithSource")

  status = clBuildProgram(
      program, 
      numDevices, 
      devices, 
      NULL,
      NULL, 
      NULL);
  if(status == CL_BUILD_PROGRAM_FAILURE) {
    size_t log_size;
    clGetProgramBuildInfo(program, devices[0],CL_PROGRAM_BUILD_LOG,0,NULL,&log_size);
    char *log = (char*) malloc(log_size);
    clGetProgramBuildInfo(program,devices[0],CL_PROGRAM_BUILD_LOG,log_size,log,NULL);
    printf("%s\n", log);
    free(log);
  }

  cl_kernel kernel = NULL;

  kernel = clCreateKernel(program, "ele_sub", &status);
  CHECK_CLSTATUS(status, "clCreateKernel")

  status  = clSetKernelArg(
      kernel, 
      0, 
      sizeof(cl_mem), 
      &in_buffer1);
  CHECK_CLSTATUS(status, "clSetKernelArg")
  status  = clSetKernelArg(
      kernel, 
      1, 
      sizeof(cl_mem), 
      &in_buffer2);
  CHECK_CLSTATUS(status, "clSetKernelArg")
  status  = clSetKernelArg(
      kernel, 
      2, 
      sizeof(cl_mem), 
      &out_buffer);
  CHECK_CLSTATUS(status, "clSetKernelArg")


  status = clEnqueueWriteBuffer(
      cmdQueue, 
      in_buffer1, 
      CL_TRUE, 
      0, 
      sizeof(in_data1),                         
      in_data1, 
      0, 
      NULL, 
      NULL);
  CHECK_CLSTATUS(status, "clEnqueueWriteBuffer")
  status = clEnqueueWriteBuffer(
      cmdQueue, 
      in_buffer2, 
      CL_TRUE, 
      0, 
      sizeof(in_data2),                         
      in_data2, 
      0, 
      NULL, 
      NULL);
  CHECK_CLSTATUS(status, "clEnqueueWriteBuffer")

  size_t globalWorkSize[] = {GX};    

  status = clEnqueueNDRangeKernel(
      cmdQueue, 
      kernel, 
      1, 
      NULL, 
      globalWorkSize, 
      NULL, 
      0, 
      NULL, 
      NULL);
  CHECK_CLSTATUS(status, "clEnqueueNDRangeKernel")

  status = clFinish(cmdQueue);
  CHECK_CLSTATUS(status, "clFinish")

  status = clEnqueueReadBuffer(
      cmdQueue, 
      out_buffer, 
      CL_TRUE, 
      0, 
      sizeof(out_data), 
      out_data, 
      0, 
      NULL, 
      NULL);
  CHECK_CLSTATUS(status, "clEnqueueReadBuffer")


  bool correct = true;
  for(int i = 0; i < GX; i++){
    if((in_data2[i] - in_data1[i]) != *(out_data + i)){
      printf("Incorrect result.\n");
      correct = false;
      break;
    }
  }
  if(correct)
    printf("Correct result.\n");

  clReleaseKernel(kernel);
  clReleaseProgram(program);
  clReleaseCommandQueue(cmdQueue);
  clReleaseMemObject(in_buffer1);
  clReleaseMemObject(in_buffer2);
  clReleaseMemObject(out_buffer);
  clReleaseContext(context);

  free(programSource);
  free(platforms);
  free(devices);

  return 0;
}
